
GAME_UPDATE_SPEED = 1/60; // how much time in seconds elapses in each game frame
GAME_RENDER_SPEED = 1/30; // how fast we actually render the screen

ANIMATION_SPEED = 8 * GAME_UPDATE_SPEED;

function timestamp() {
	return Date.now() / 1000;
}

BOARD_OFFSET_X =103;
BOARD_OFFSET_Y = 120;
BOARD_TILES_W = 6;
BOARD_TILES_H = 6;
TILE_SIZE = 130;
BOARD_WIDTH = 1000;
BOARD_HEIGHT = 1020;
SELECTION_OFFSET = 10;
SELECTION_START_IDX = 5;

function isKing(num) { return (num % 6) == 0; }
function isQueen(num) { return (num % 6) == 1; }
function isRook(num) { return (num % 6) == 2; }
function isKnight(num) { return (num % 6) == 3; }
function isBishop(num) { return (num % 6) == 4; }
function isPawn(num) { return (num % 6) == 5; }
function isWhite(num) { return num < 6; }

g_Board = [

  -1, 2, 0, 1, 2,-1,
  -1, 3, 4, 4, 3,-1,
  -1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,
  -1, 9,10,10, 9,-1,
  -1, 8, 6, 7, 8,-1,

];

if (false) { // test for checkmate/stalemate easier
	
	g_Board = [
		 0,-1, 2,-1, 2,-1,
		-1, 2,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,
		-1,-1,-1,-1,-1,-1,
		-1,-1,-1, 6,-1,-1,
	]

}

GAMESTATE_GAME_ON = "game_on";
GAMESTATE_STALEMATE = "Stalemate!";
GAMESTATE_CHECKMATE = "Won by Checkmate!";
GAMESTATE_KINGSLAYER = "Won by Kingslaying!"
GAMESTATE_IN_CHECK = "Checked!";

g_GameState = GAMESTATE_GAME_ON

g_SelectedIdx = -1;
g_SelectedMoves = null;
g_IsWhiteTurn = true;

g_QueuedMoveFrom = -1;
g_QueuedMoveTo = -1;

// this will be overwritten from our html
g_CanvasWidth = 1000;
g_CanvasHeight = 1000;
g_NeedsPaint = true;

// some math utility functions
function rgba(r, g, b, a) {

	return 'rgba(' + Math.floor(r) + ','+ Math.floor(g) + ',' + Math.floor(b) + ',' + a + ')';
}

function square(x) {

	return x*x;
}

function distanceSq(x1,y1, x2,y2) {

	return square(x1-x2) + square(y1-y2);
}

// image loading support
g_ImageCache = {};
g_SoundCache = {};
g_AssetsToLoad = [];

function loadImage(path) {

	if (path in g_ImageCache) {
		return g_ImageCache[path];
	}

	g_AssetsToLoad.push(path);

	var image = new Image();
	image.onload = function () {
		g_AssetsToLoad.splice(g_AssetsToLoad.indexOf(path), 1);
	};
	image.src = path;

	g_ImageCache[path] = image;
	return image;
}

function playAudio(obj) {

	try {
		obj.audio[obj.index].play();

	} catch (err) {

	}

	obj.index = (obj.index+1)%obj.audio.length;
}

function loadAudio(path, count=1) {

	if (path in g_SoundCache) {
		return g_SoundCache[path];
	}

	result = new Object();
	result.audio = []
	result.index = 0;

	for (var n = 0; n < count; ++n){
		var a = new Audio(path);
		a.load();
		result.audio.push(a);
	}

	g_SoundCache[path] = result;

	return result;
}

function getKingIdx(board, is_white) {

	for (var n = 0; n < board.length; ++n) {
		if (isKing(board[n]) && isWhite(board[n]) == is_white) {
			return n;
		}
	}

	// we can get here if the kings are killed (side effect of delays)
	return -1;
}

function isKingInCheck(board, is_white) {

	var king_idx = getKingIdx(board, is_white);

	for (var n = 0; n < board.length; ++n) {
		if (board[n] != -1 && isWhite(board[n]) != is_white) {

			var moves = getValidMoves(board, n, false);

			if (moves.indexOf(king_idx) != -1)
				return true;
		}
	}

	return false;
}

function getGameState(board, is_white) {

	if (getKingIdx(board, true) == -1 ||
		getKingIdx(board, false) == -1) {

		return GAMESTATE_KINGSLAYER;
	}

	// check to see if we have *any*GAMESTATE_KINGSLAYER moves
	for (var n = 0; n < board.length; ++n) {
		if (board[n] != -1 && isWhite(board[n]) == is_white) {

			if (getValidMoves(board, n, true).length != 0)
			{
				if (isKingInCheck(board, is_white))
					return GAMESTATE_IN_CHECK;

				return GAMESTATE_GAME_ON;
			}
		}
	}

	// test to see if we are in stalemate or checkmate
	if (isKingInCheck(board, is_white)) {

		return GAMESTATE_CHECKMATE;
	}

	return GAMESTATE_STALEMATE;
}

function getValidMoves(board, from_idx, test_for_check) {

	var result = []
	var from_piece = board[from_idx];
	var from_piece_is_white = isWhite(from_piece);
	var from_x = Math.floor(from_idx % BOARD_TILES_W);
	var from_y = Math.floor(from_idx / BOARD_TILES_W);

	if (from_piece == -1)
		return result;

	function add(off_x, off_y) {

		var to_x = from_x + off_x;
		var to_y = from_y + off_y;

		if (to_x < 0 || to_y < 0 || to_x >= BOARD_TILES_W || to_y >= BOARD_TILES_H)
			return false;

		var to_idx = to_x + to_y*BOARD_TILES_W;

		var to_piece = board[to_idx];

		if (to_piece != -1 && from_piece_is_white == isWhite(to_piece))
		{
			return false;
		}

		if (test_for_check) {
			var board_copy = board.slice();

			// only allow movement when near king
			if (isNearKing(board, from_idx))
				makeMove(board_copy, from_idx, to_idx);

			if (isKingInCheck(board_copy, from_piece_is_white)) {

				return false;
			}
		}

		result.push(to_idx);

		if (to_piece != -1 && from_piece_is_white != isWhite(to_piece))
			return false;

		return true;
	}


	if (isKing(from_piece)) {

		add(-1,-1);
		add(+1,-1);
		add(-1,+1);
		add(+1,+1);

		add( 0,-1);
		add( 0,+1);
		add(-1, 0);
		add(+1, 0);
	}
	else if (isKnight(from_piece)) {

		add(-1,-2);
		add(-2,-1);

		add(+1,-2);
		add(+2,-1);

		add(-1,+2);
		add(-2,+1);

		add(+1, +2);
		add(+2, +1);

	}
	else {

		if (isRook(from_piece) || isQueen(from_piece)) {

			for (var n = 1; add(-n,0); ++n) {}
			for (var n = 1; add(+n,0); ++n) {}
			for (var n = 1; add(0,-n); ++n) {}
			for (var n = 1; add(0,+n); ++n) {}

		}

		if (isBishop(from_piece) || isQueen(from_piece)) {

			for (var n = 1; add(-n, -n); n++) {}
			for (var n = 1; add(+n, -n); n++) {}
			for (var n = 1; add(-n, +n); n++) {}
			for (var n = 1; add(+n, +n); n++) {}

		}
	}

	return result;
}

function makeMove(board, from_idx, to_idx) {

	var existing = board[from_idx];
	board[from_idx] = -1;
	board[to_idx] = existing;
}

function performDelayedMove(board, from_idx, to_idx, is_white_turn) {

	// slightly different logic for delayed moves, 
	var from_piece = board[from_idx];

	if (isWhite(from_piece) != is_white_turn) // should never hit, but this logic is hurting my brain so just in case
		return; 

	if (isKnight(from_piece) || isKing(from_piece)) {

		makeMove(board, from_idx, to_idx);
	}
	else {

		// move rooks/bishops until they take a piece

		var from_x = Math.floor(from_idx % 6);
		var from_y = Math.floor(from_idx / 6);

		var to_x = Math.floor(to_idx % 6);
		var to_y = Math.floor(to_idx / 6);

		var off_x = 0;
		var off_y = 0;

		if (to_x < from_x) off_x = -1;
		else if (to_x > from_x) off_x = 1;
		 
		if (to_y < from_y) off_y = -1;
		else if (to_y > from_y) off_y = 1;

		var x = from_x+off_x;
		var y = from_y+off_y;

		while (y != to_y && x != to_x) {

			var current_idx = y*BOARD_TILES_W+x;

			if (board[current_idx] != -1) {

				// move to this space... even if we stomp our own piece (should be hard to do)
				makeMove(board, from_idx, current_idx);
				return;
			}

			x += off_x;
			y += off_y;
		}

		// if we reached here nothing blocked us
		makeMove(board, from_idx, to_idx);
	}
}

function drawSelection(ctx) {

	if (g_SelectedIdx != -1) {

		var selected_idx = SELECTION_START_IDX + (g_IsWhiteTurn ? 0 : 6);

		for (var n = 0; n < g_SelectedMoves.length; ++n) {

			var x = Math.floor(g_SelectedMoves[n] % BOARD_TILES_W);
			var y = Math.floor(g_SelectedMoves[n] / BOARD_TILES_W);

			ctx.drawImage(g_TilesImage, 
				(selected_idx%6)*TILE_SIZE, Math.floor(selected_idx/6)*TILE_SIZE, TILE_SIZE, TILE_SIZE,
				BOARD_OFFSET_X+x*TILE_SIZE, BOARD_OFFSET_Y+y*TILE_SIZE, TILE_SIZE, TILE_SIZE
				);
		}
	}

	// also draw queued moves
	if (g_QueuedMoveFrom != -1) {

		var from_x = Math.floor(g_QueuedMoveFrom % 6);
		var from_y = Math.floor(g_QueuedMoveFrom / 6);

		var to_x = Math.floor(g_QueuedMoveTo % 6);
		var to_y = Math.floor(g_QueuedMoveTo / 6);

		var selected_idx = SELECTION_START_IDX + (g_IsWhiteTurn ? 6 : 0);

		// always draw from/to idx
		ctx.drawImage(g_TilesImage, 
			(selected_idx%6)*TILE_SIZE, Math.floor(selected_idx/6)*TILE_SIZE, TILE_SIZE, TILE_SIZE,
			BOARD_OFFSET_X+from_x*TILE_SIZE, BOARD_OFFSET_Y+from_y*TILE_SIZE, TILE_SIZE, TILE_SIZE
			);

		ctx.drawImage(g_TilesImage, 
			(selected_idx%6)*TILE_SIZE, Math.floor(selected_idx/6)*TILE_SIZE, TILE_SIZE, TILE_SIZE,
			BOARD_OFFSET_X+to_x*TILE_SIZE, BOARD_OFFSET_Y+to_y*TILE_SIZE, TILE_SIZE, TILE_SIZE
			);


		// draw arc for non-knights
		if (!isKnight(g_Board[g_QueuedMoveFrom])) {

			var off_x = 0;
			var off_y = 0;

			if (to_x < from_x) off_x = -1;
			else if (to_x > from_x) off_x = 1;
			 
			if (to_y < from_y) off_y = -1;
			else if (to_y > from_y) off_y = 1;

			var x = from_x+off_x;
			var y = from_y+off_y;

			while (y != to_y && x != to_x) {
				ctx.drawImage(g_TilesImage, 
					(selected_idx%6)*TILE_SIZE, Math.floor(selected_idx/6)*TILE_SIZE, TILE_SIZE, TILE_SIZE,
					BOARD_OFFSET_X+x*TILE_SIZE, BOARD_OFFSET_Y+y*TILE_SIZE, TILE_SIZE, TILE_SIZE
					);

				x += off_x;
				y += off_y;
			}
		}
	}
}

function drawPieces(ctx) {

	// go over all of our current tiles and draw them
	for (var y = 0; y < BOARD_TILES_H; ++y) {

		for (var x = 0; x < BOARD_TILES_W; ++x) {

			var piece_idx = y*6+x;
			var piece = g_Board[piece_idx];

			if (piece != -1) {

				var selection_offset = 0;
				if (g_SelectedIdx == piece_idx)
					selection_offset = SELECTION_OFFSET;

				ctx.drawImage(g_TilesImage, 
					(piece%6)*TILE_SIZE, Math.floor(piece/6)*TILE_SIZE, TILE_SIZE, TILE_SIZE,
					BOARD_OFFSET_X+x*TILE_SIZE, BOARD_OFFSET_Y+y*TILE_SIZE-selection_offset, TILE_SIZE, TILE_SIZE
					);

				if (isNearKing(g_Board, piece_idx)) {

					ctx.fillStyle = '#0F0';
				}
				else if (g_QueuedMoveFrom == piece_idx) {
					ctx.fillStyle = '#FF0';
				}
				else {
					ctx.fillStyle = '#F00';
				}				ctx.strokeStyle = '#000';
				ctx.lineWidth = 3;


				ctx.strokeStyle = '#000';
				ctx.lineWidth = 3;
				ctx.beginPath();
				ctx.arc(BOARD_OFFSET_X+x*TILE_SIZE + TILE_SIZE*1/2, BOARD_OFFSET_Y+y*TILE_SIZE + TILE_SIZE*7/8, 8, 0, 2 * Math.PI);
				ctx.fill();
				ctx.stroke();

			}
		}
	}
}

function update() {

}

function drawLoadScreen() {

	var cvs = document.getElementById("cvs");
	var ctx = cvs.getContext('2d');

	ctx.fillStyle = '#000';
	ctx.fillRect(0, 0, g_CanvasWidth, g_CanvasHeight);

	ctx.fillStyle = '#F0F';
	ctx.font = '24px Pacifico serif';
	var label = "Loading";

	var w = ctx.measureText(label).width;

	ctx.fillText(label, (cvs.width-w)/2, cvs.height/2+10);
}

function drawGameState(ctx) {

	if (g_GameState == GAMESTATE_GAME_ON)
		return;

	var color = g_IsWhiteTurn;

	if (g_GameState == GAMESTATE_KINGSLAYER)
		color = !color;

	if (color)
		ctx.fillStyle = '#F0F';
	else
		ctx.fillStyle = '#0FF';

	if (g_GameState == GAMESTATE_STALEMATE)
		ctx.fillStyle = '#FFF';

	ctx.strokeStyle = "black";
	ctx.lineWidth = 8;

	ctx.font = '24px Pacifico serif';

	var w = ctx.measureText(g_GameState).width;

	var x = (cvs.width-w)/2;
	var y = cvs.height/2+10;

	ctx.strokeText(g_GameState, x, y);
	ctx.fillText(g_GameState, x, y);

}

function draw() {

	// skip drawing when nothing changed
	if (g_NeedsPaint == false)
		return;

	g_NeedsPaint = false;

	var cvs = document.getElementById("cvs");
	var ctx = cvs.getContext('2d');

	ctx.drawImage(g_BackgroundImage, 
		(g_IsWhiteTurn ? 0 : 1) * BOARD_WIDTH, 0, BOARD_WIDTH, BOARD_HEIGHT,
		0, 0, BOARD_WIDTH, BOARD_HEIGHT
		);

	drawSelection(ctx);
	drawPieces(ctx);
	drawGameState(ctx);
}

function mainLoop() {

	if (g_AssetsToLoad.length > 0) {

		drawLoadScreen();

		g_CalculatedTime = timestamp();

	} else {

		/*
		// this updates our fixed game loop multiple times until we reach real time
		while (g_CalculatedTime < timestamp()) {
			
			update();

			g_CalculatedTime += GAME_UPDATE_SPEED;
		}
		*/

		draw();
	}
}

//function addPulser(x, y, color, spawnDelay, radius, waveSpeed, initialTime = 0) {
window.onload = function(e) {

	var cvs = document.getElementById("cvs");

	g_CanvasWidth = cvs.width;
	g_CanvasHeight = cvs.height;

	g_BackgroundImage = loadImage("background.png");
	g_TilesImage = loadImage("tiles.png");

	setInterval(mainLoop, GAME_RENDER_SPEED);
};

function isNearKing(board, idx) {

	if (isKing(board[idx])) {
		return true;
	}

	var px = Math.floor(idx%BOARD_TILES_W);
	var py = Math.floor(idx/BOARD_TILES_H);

	var king_idx = getKingIdx(board, isWhite(board[idx]));
	var kx = Math.floor(king_idx%BOARD_TILES_W);
	var ky = Math.floor(king_idx/BOARD_TILES_H);

	return (Math.abs(px-kx) <= 1 && Math.abs(py-ky) <= 1);

}

window.onclick = function(e) {

	if (g_GameState != GAMESTATE_GAME_ON && g_GameState != GAMESTATE_IN_CHECK)
		return;

	var tx = Math.floor((e.offsetX - BOARD_OFFSET_X) / TILE_SIZE);
	var ty = Math.floor((e.offsetY - BOARD_OFFSET_Y) / TILE_SIZE);

	var new_idx = tx + ty * 6;

	if (tx >= 0 && tx < BOARD_TILES_W && ty >= 0 && ty < BOARD_TILES_H) {

		if (g_SelectedIdx != -1) {

			if (g_SelectedMoves.indexOf(new_idx) != -1) {

				var move_from = g_QueuedMoveFrom;
				var move_to = g_QueuedMoveTo;

				if (isNearKing(g_Board, g_SelectedIdx)) {

					// near king moves immediately
					makeMove(g_Board, g_SelectedIdx, new_idx);

					g_QueuedMoveFrom = -1;
					g_QueuedMoveTo = -1;
				}
				else {

					g_QueuedMoveFrom = g_SelectedIdx;
					g_QueuedMoveTo = new_idx;
				}

				g_SelectedIdx = -1;
				g_SelectedMoves = null;
				g_NeedsPaint = true;

				g_IsWhiteTurn = !g_IsWhiteTurn;

				// play black moves

				if (move_from != -1) {

					performDelayedMove(g_Board, move_from, move_to, g_IsWhiteTurn);

					// if the delayed move wiped out or move, cancel it
					if (isWhite(g_Board[g_QueuedMoveFrom]) == g_IsWhiteTurn) {

						g_QueuedMoveFrom = -1;
						g_QueuedMoveTo = -1;
					}
				}

				g_GameState = getGameState(g_Board, g_IsWhiteTurn);
				return;
			}
		}

		if (g_Board[new_idx] != -1) {

			if (new_idx != g_SelectedIdx && isWhite(g_Board[new_idx]) == g_IsWhiteTurn) {

				g_SelectedIdx = new_idx;

				g_SelectedMoves = getValidMoves(g_Board, g_SelectedIdx, true);

				g_NeedsPaint = true;
			} 
			else {

				if (g_SelectedIdx != -1) {

					g_SelectedIdx = -1;
					g_SelectedMoves = null;
					g_NeedsPaint = true;
				}
			}
		}
		else
		{
			if (g_SelectedIdx != -1) {

				g_SelectedIdx = -1;
				g_SelectedMoves = null;
				g_NeedsPaint = true;
			}
		}
	}
	else
	{
		if (g_SelectedIdx != -1) {

			g_NeedsPaint = true;
		}

		g_SelectedMoves = null;
		g_SelectedIdx = -1;
	}
}

